#
# tcllian::load rss.tcl
#
# RSS 1.0 feed:
#
#   ::rss::add {Valve} http://steampowered.com/rss.xml 5 10
#
# RSS 2.0 feed:
#
#   ::rss::add {MSDN Subscriptions} http://msdn.microsoft.com/subscriptions/rss.xml 5 10
#
# Atom 0.3 feed:
#
#   ::rss::add {Gizmodo} http://www.gizmodo.com/atom.xml 5 10
#
# Start RSS feed reader:
#
#   ::rss::start
#

package require tdom

namespace eval ::rss {
    variable feeds
    variable xml

    proc start {} {
        refresh
    }

    proc stop {} {
        variable feeds

        foreach name [array names feeds] {
            remove $name
        }
    }

    #
    # name: display name of the feed
    # url: URL to the RSS XML data
    # frequency: how many minutes between updates
    # limit: how many items to display (0 = unlimited)
    #
    proc add {name url frequency limit} {
        variable feeds

        set feeds($name) [list -1 $url $frequency $limit {}]
    }

    proc remove {name} {
        variable feeds
        variable xml

        foreach {id url frequency limit after} $feeds($name) break
        catch {after cancel $after}
        catch {unset xml($name)}
        set feeds($name) [list -1 $url $frequency $limit {}]
        trillian listRemoveEntry $id
    }

    proc purge {name} {
        variable feeds

        remove $name
        unset feeds($name)
    }
    
    proc refresh {} {
        variable feeds

        foreach feed [array names feeds] {
            refreshFeed $feed
        }
    }

    proc refreshFeed {name} {
        variable feeds
        variable xml

        foreach {id url frequency limit after} $feeds($name) break
        if {$id == -1} {
            set id [trillian listAddEntry \
                group 1 section_id $::tcllian::section_id \
                previous_id -2 expanded 1 \
                real_name $name text $name]
            set feeds($name) [list $id $url $frequency $limit $after]
        } else {
            trillian listClearEntry $id
        }
        trillian httpFileRequest $url [list ::rss::httpCallback $name]

        if {[lsearch [after info] $after] < 0} {
            set after [after [expr {$frequency * 60 * 1000}] \
                [list ::rss::refreshFeed $name]]
            set feeds($name) [list $id $url $frequency $limit $after]
        }
    }

    proc httpCallback {name event result} {
        variable feeds
        variable xml

        ## tDOM doesn't like the "windows-1252" encoding, so lets
        ## convert to UTF-8 and fake it.
        if {[regexp {<\?xml\s.*\sencoding="windows-1252".*?\?>} $result]} {
            set result [encoding convertfrom cp1252 $result]
            regsub {(<\?xml\s.*\s)encoding="windows-1252"(.*?\?>)} $result \
                    {\1encoding="utf-8"\2} result
        }

        foreach {id url frequency limit after} $feeds($name) break
        set xml($name) $result
        if {[catch {dom parse $result dom} err]} {
            #trillian systraySetAlert . "RSS error: $err"
            tcllian messagebox $err "RSS error"
            return
        }
        set result [catch {
            $dom documentElement doc
            switch -exact -- [$doc nodeName] {
                rss {
                    # RSS 0.9x, RSS 2.0
                    set itemXpath	{/rss/channel/item}
                    set titleXpath	{string(title)}
                    set linkXpath	{string(link)}
                    set descXpath	{string(description)}
                }
                rdf:RDF {
                    # RSS 1.0
                    $doc setAttributeNS "" xmlns:null [$doc getAttribute xmlns]
                    set itemXpath	{/rdf:RDF/null:item}
                    set titleXpath	{string(null:title)}
                    set linkXpath	{string(null:link)}
                    set descXpath	{string(content:encoded)}
                }
                feed {
                    # Atom 0.3
                    $doc setAttributeNS "" xmlns:null [$doc getAttribute xmlns]
                    set itemXpath	{/null:feed/null:entry}
                    set titleXpath	{string(null:title)}
                    set linkXpath	{string(null:link/@href)}
                    set descXpath	{string(null:summary)}
                }
            }
            set count 0
            foreach node [$doc selectNodes $itemXpath] {
                set title [$node selectNodes $titleXpath]
                set link [$node selectNodes $linkXpath]
                set description [$node selectNodes $descXpath]
                if {[string length $link]} {
                    trillian listAddEntry \
                        section_id $::tcllian::section_id parent_id $id \
                        real_name $title text $title tooltip $description \
                        callback [list trillian browserOpen $link]
                } else {
                    trillian listAddEntry \
                        section_id $::tcllian::section_id parent_id $id \
                        real_name $title text $title tooltip $description
                }
                incr count
                if {$limit && $count == $limit} {
                    break
                }
            }
        } err]
        $dom delete
        if {$result} {
            trillian systraySetAlert . "RSS Error: $err"
        }
    }
}

